@@ -2,4 +2,7 @@
 -- Back then, there was no state column
 SELECT COUNT(*) as active_connections
   FROM pg_stat_activity
- WHERE current_query NOT LIKE '<IDLE>%%';
+ WHERE current_query NOT LIKE '<IDLE>%%'
+ AND CASE WHEN %(dbname_filter)s IS NULL THEN true
+    ELSE datname ~* %(dbname_filter)s
+    END;
