@@ -10,107 +10,23 @@ from typing import (
     Mapping,
     MutableSet,
     Optional,
+    overload,
     Sequence,
     Tuple,
-    Type,
     TypeVar,
     Union,
 )
 
 import attr
 import psutil
+from attr import validators
 
-from . import colors, utils
+from . import compat, colors, utils
 
 
 T = TypeVar("T")
 
 
-class Deserializable:
-    """Mixin class adding deserialization support.
-
-    >>> @attr.s(auto_attribs=True)
-    ... class Point(Deserializable):
-    ...     x: int
-    ...     y: int
-    ...     label: Optional[str] = None
-
-    >>> data = {"x": 1, "y": 1, "label": "a"}
-    >>> Point.deserialize(data)
-    Point(x=1, y=1, label='a')
-
-    >>> @attr.s(auto_attribs=True)
-    ... class Line(Deserializable):
-    ...     start: Point
-    ...     end: Point
-    ...     color: str = "black"
-    ...     label: Optional[str] = None
-
-    >>> data = {"start": {"x": 1, "y": 1}, "end": {"x": 2, "y": 2, "label": "p2"}}
-    >>> Line.deserialize(data)
-    Line(start=Point(x=1, y=1, label=None), end=Point(x=2, y=2, label='p2'), color='black', label=None)
-
-    >>> data = {"start": {"x": 1, "y": 1}, "end": {"x": 2, "y": 2}, "colour": "red"}
-    >>> Line.deserialize(data)
-    Traceback (most recent call last):
-        ...
-    ValueError: unknown field(s): colour
-
-    >>> data = {"start": {"x": 1, "y": 1}}
-    >>> Line.deserialize(data)
-    Traceback (most recent call last):
-        ...
-    ValueError: missing required field 'end'
-
-    >>> data = {"start": {"x": 1, "y": 1}, "end": {"x": 2, "y": 2}, "color": (255, 5, 2)}
-    >>> Line.deserialize(data)
-    Traceback (most recent call last):
-        ...
-    TypeError: invalid type for field 'color', expecting <class 'str'>
-    """
-
-    @classmethod
-    def deserialize(cls: Type[T], data: Mapping[str, Any]) -> T:
-        args = {}
-        for field in attr.fields(cls):
-            name = field.name
-            try:
-                value = data[name]
-            except KeyError:
-                if field.default != attr.NOTHING:
-                    continue
-                raise ValueError(f"missing required field '{name}'") from None
-            else:
-                try:
-                    deserializer = getattr(field.type, "deserialize")
-                except AttributeError:
-                    assert field.type is not None, "fields should be typed"
-                    try:
-                        is_subtype = isinstance(value, field.type)
-                    except TypeError:
-                        # This might happen for Union types (e.g.
-                        # Optional[X]), we assume the type is okay waiting for
-                        # a better strategy.
-                        pass
-                    else:
-                        if not is_subtype:
-                            raise TypeError(
-                                f"invalid type for field '{name}', expecting {field.type}"
-                            ) from None
-                else:
-                    value = deserializer(value)
-
-                args[name] = value
-
-        unknown = set(data) - set(args)
-        if unknown:
-            raise ValueError(
-                f"unknown field(s): {', '.join(sorted(unknown))}"
-            ) from None
-
-        return cls(**args)  # type: ignore[call-arg]
-
-
 E = TypeVar("E", bound=enum.IntEnum)
 
 
@@ -123,29 +39,19 @@ def enum_next(e: E) -> E:
     ...     summer = 3
     ...     autumn = 4
 
-    >>> enum_next(Seasons.winter)
-    <Seasons.spring: 2>
-    >>> enum_next(Seasons.spring)
-    <Seasons.summer: 3>
-    >>> enum_next(Seasons.autumn)
-    <Seasons.winter: 1>
+    >>> enum_next(Seasons.winter).name
+    'spring'
+    >>> enum_next(Seasons.spring).name
+    'summer'
+    >>> enum_next(Seasons.autumn).name
+    'winter'
     """
     return e.__class__((e.value % max(e.__class__)) + 1)
 
 
 @enum.unique
 class Flag(enum.IntFlag):
-    """Column flag.
-
-    >>> f = Flag(3)
-    >>> f
-    <Flag.APPNAME|DATABASE: 3>
-    >>> f | Flag.CLIENT
-    <Flag.CLIENT|APPNAME|DATABASE: 7>
-    >>> f ^= Flag.APPNAME
-    >>> f
-    <Flag.DATABASE: 1>
-    """
+    """Column flag."""
 
     DATABASE = 1
     APPNAME = 2
@@ -185,31 +91,7 @@ class Flag(enum.IntFlag):
         nowrite: bool,
         **kwargs: Any,
     ) -> "Flag":
-        """Build a Flag value from command line options.
-
-        >>> options = {
-        ... 'noappname': False,
-        ... 'noclient': False,
-        ... 'nocpu': False,
-        ... 'nodb': False,
-        ... 'nomem': False,
-        ... 'nopid': False,
-        ... 'noread': False,
-        ... 'notime': False,
-        ... 'nouser': False,
-        ... 'nowait': False,
-        ... 'nowrite': False,
-        ... }
-        >>> Flag.from_options(is_local=True, **options)
-        <Flag.PID|IOWAIT|MODE|TYPE|RELATION|WAIT|TIME|WRITE|READ|MEM|CPU|USER|CLIENT|APPNAME|DATABASE: 32767>
-        >>> Flag.from_options(is_local=False, **options)
-        <Flag.PID|MODE|TYPE|RELATION|WAIT|TIME|USER|CLIENT|APPNAME|DATABASE: 24335>
-        >>> options['nodb'] = True
-        >>> options['notime'] = True
-        >>> options['nopid'] = True
-        >>> Flag.from_options(is_local=False, **options)
-        <Flag.MODE|TYPE|RELATION|WAIT|USER|CLIENT|APPNAME: 7694>
-        """
+        """Build a Flag value from command line options."""
         flag = cls.all()
         if nodb:
             flag ^= cls.DATABASE
@@ -248,6 +130,32 @@ class Flag(enum.IntFlag):
         return flag
 
 
+@attr.s(auto_attribs=True, frozen=True, slots=True)
+class Filters:
+    dbname: Optional[str] = None
+
+    @classmethod
+    def from_options(cls, filters: Sequence[str]) -> "Filters":
+        fields = compat.fields_dict(cls)
+        attrs = {}
+        for f in filters:
+            try:
+                fname, regex = f.split(":", 1)
+            except ValueError:
+                raise ValueError(f"malformatted filter value '{f}'")
+            if not regex:
+                raise ValueError(f"empty regex in filter '{f}'")
+            if fname in attrs:
+                raise ValueError(f"got multiple filters '{fname}'")
+            if fname not in fields:
+                raise ValueError(f"unknown filter '{fname}'")
+            attrs[fname] = regex
+        return cls(**attrs)
+
+
+NO_FILTER = Filters()
+
+
 class SortKey(enum.Enum):
     cpu = enum.auto()
     mem = enum.auto()
@@ -296,7 +204,8 @@ _color_key_marker = f"{id(object())}"
 class Column:
     """A column in stats table.
 
-    >>> c = Column("pid", "PID", "%-6s", True, SortKey.cpu, max_width=6,
+    >>> c = Column("pid", "PID", mandatory=True, sort_key=SortKey.cpu,
+    ...            min_width=6, max_width=6,
     ...            transform=lambda v: str(v)[::-1])
     >>> c.title_render()
     'PID   '
@@ -310,45 +219,55 @@ class Column:
     '876543'
     >>> c.color_key
     'pid'
+
+    >>> c = attr.evolve(c, justify="right", min_width=4, max_width=5)
+    >>> c.title_render()
+    ' PID'
+    >>> c.render('7654321')
+    '12345'
+    >>> c.render('21')
+    '  12'
     """
 
     key: str = attr.ib(repr=False)
     name: str
-    template_h: str = attr.ib()
     mandatory: bool = False
     sort_key: Optional[SortKey] = None
+    min_width: int = attr.ib(default=0, repr=False)
     max_width: Optional[int] = attr.ib(default=None, repr=False)
+    justify: str = attr.ib(
+        "left", validator=validators.in_(["left", "center", "right"])
+    )
     transform: Callable[[Any], str] = attr.ib(default=str, repr=False)
     color_key: Union[str, Callable[[Any], str]] = attr.ib(
         default=_color_key_marker, repr=False
     )
 
-    @template_h.validator
-    def _template_h_is_a_format_string_(self, attribute: Any, value: str) -> None:
-        """Validate template_h attribute.
-
-        >>> Column("k", "a", "b%%aa")
-        Traceback (most recent call last):
-            ...
-        ValueError: template_h must be a format string with one placeholder
-        >>> Column("k", "a", "baad")
-        Traceback (most recent call last):
-            ...
-        ValueError: template_h must be a format string with one placeholder
-        >>> Column("k", "a", "%s is good")  # doctest: +ELLIPSIS
-        Column(name='a', template_h='%s is good', ...)
-        """
-        if value.count("%") != 1:
-            raise ValueError(
-                f"{attribute.name} must be a format string with one placeholder"
-            )
+    _justify: Callable[[str], str] = attr.ib(init=False)
 
     def __attrs_post_init__(self) -> None:
         if self.color_key == _color_key_marker:
             object.__setattr__(self, "color_key", self.key)
 
+        if self.justify == "left":
+
+            def _justify(value: str) -> str:
+                return value.ljust(self.min_width)[: self.max_width]
+
+        elif self.justify == "right":
+
+            def _justify(value: str) -> str:
+                return value.rjust(self.min_width)[: self.max_width]
+
+        elif self.justify == "center":
+
+            def _justify(value: str) -> str:
+                return value.center(self.min_width)[: self.max_width]
+
+        object.__setattr__(self, "_justify", _justify)
+
     def title_render(self) -> str:
-        return self.template_h % self.name
+        return self._justify(self.name)
 
     def title_color(self, sort_by: SortKey) -> str:
         if self.sort_key == sort_by:
@@ -356,7 +275,7 @@ class Column:
         return "green"
 
     def render(self, value: Any) -> str:
-        return self.template_h % self.transform(value)[: self.max_width]
+        return self._justify(self.transform(value))
 
     def color(self, value: Any) -> str:
         if callable(self.color_key):
@@ -373,7 +292,7 @@ class UI:
     duration_mode: DurationMode = attr.ib(
         default=DurationMode.query, converter=DurationMode
     )
-    verbose_mode: QueryDisplayMode = attr.ib(
+    query_display_mode: QueryDisplayMode = attr.ib(
         default=QueryDisplayMode.default(), converter=QueryDisplayMode
     )
     sort_key: SortKey = attr.ib(default=SortKey.default(), converter=SortKey)
@@ -388,42 +307,45 @@ class UI:
         flag: Flag = Flag.all(),
         *,
         max_db_length: int = 16,
+        filters: Filters = NO_FILTER,
         **kwargs: Any,
     ) -> "UI":
         possible_columns: Dict[str, Column] = {}
 
-        def add_column(key: str, **kwargs: Any) -> None:
+        def add_column(key: str, name: str, **kwargs: Any) -> None:
             assert key not in possible_columns, f"duplicated key {key}"
-            possible_columns[key] = Column(key, **kwargs)
+            possible_columns[key] = Column(key=key, name=name, **kwargs)
 
         if Flag.APPNAME & flag:
             add_column(
                 key="application_name",
                 name="APP",
-                template_h="%16s ",
+                min_width=16,
                 max_width=16,
+                justify="right",
             )
         if Flag.CLIENT & flag:
             add_column(
                 key="client",
                 name="CLIENT",
-                template_h="%16s ",
+                min_width=16,
                 max_width=16,
+                justify="right",
             )
         if Flag.CPU & flag:
             add_column(
                 key="cpu",
                 name="CPU%",
-                template_h="%6s ",
+                min_width=6,
                 sort_key=SortKey.cpu,
             )
         if Flag.DATABASE & flag:
             add_column(
                 key="database",
-                name="DATABASE",
-                template_h=f"%-{max_db_length}s ",
+                name="DATABASE(*)" if filters.dbname else "DATABASE",
+                min_width=max_db_length,
                 transform=functools.lru_cache()(
-                    functools.partial(utils.ellipsis, width=16)
+                    lambda v: utils.ellipsis(v, width=16) if v else "",
                 ),
                 sort_key=None,
             )
@@ -431,7 +353,7 @@ class UI:
             add_column(
                 key="io_wait",
                 name="IOW",
-                template_h="%4s ",
+                min_width=4,
                 transform=utils.yn,
                 color_key=colors.wait,
             )
@@ -439,7 +361,7 @@ class UI:
             add_column(
                 key="mem",
                 name="MEM%",
-                template_h="%4s ",
+                min_width=4,
                 sort_key=SortKey.mem,
                 transform=lambda v: str(round(v, 1)),
             )
@@ -447,18 +369,27 @@ class UI:
             add_column(
                 key="mode",
                 name="MODE",
-                template_h="%16s ",
+                min_width=16,
                 max_width=16,
+                justify="right",
                 color_key=colors.lock_mode,
             )
         if Flag.PID & flag:
-            add_column(key="pid", name="PID", template_h="%-6s ")
-        add_column(key="query", name="Query", template_h=" %2s")
+            add_column(
+                key="pid",
+                name="PID",
+                min_width=6,
+            )
+        add_column(
+            key="query",
+            name="Query",
+            min_width=2,
+        )
         if Flag.READ & flag:
             add_column(
                 key="read",
                 name="READ/s",
-                template_h="%8s ",
+                min_width=8,
                 sort_key=SortKey.read,
                 transform=utils.naturalsize,
             )
@@ -466,13 +397,15 @@ class UI:
             add_column(
                 key="relation",
                 name="RELATION",
-                template_h="%9s ",
+                min_width=9,
                 max_width=9,
+                justify="right",
             )
         add_column(
             key="state",
             name="state",
-            template_h=" %17s  ",
+            min_width=17,
+            justify="right",
             transform=utils.short_state,
             color_key=colors.short_state,
         )
@@ -480,28 +413,43 @@ class UI:
             add_column(
                 key="duration",
                 name="TIME+",
-                template_h="%9s ",
+                min_width=9,
+                justify="right",
                 sort_key=SortKey.duration,
                 transform=lambda v: utils.format_duration(v)[0],
                 color_key=lambda v: utils.format_duration(v)[1],
             )
         if Flag.TYPE & flag:
-            add_column(key="type", name="TYPE", template_h="%16s ", max_width=16)
+            add_column(
+                key="type",
+                name="TYPE",
+                min_width=16,
+                max_width=16,
+                justify="right",
+            )
         if Flag.USER & flag:
-            add_column(key="user", name="USER", template_h="%16s ", max_width=16)
+            add_column(
+                key="user",
+                name="USER",
+                min_width=16,
+                max_width=16,
+                justify="right",
+            )
         if Flag.WAIT & flag:
             add_column(
                 key="wait",
-                name="W",
-                template_h="%2s ",
-                transform=utils.yn,
+                name="Waiting",
+                min_width=16,
+                max_width=16,
+                justify="right",
+                transform=utils.wait_status,
                 color_key=colors.wait,
             )
         if Flag.WRITE & flag:
             add_column(
                 key="write",
                 name="WRITE/s",
-                template_h="%8s ",
+                min_width=8,
                 sort_key=SortKey.write,
                 transform=utils.naturalsize,
             )
@@ -546,6 +494,7 @@ class UI:
                 "type",
                 "mode",
                 "duration",
+                "wait",
                 "state",
                 "query",
             ],
@@ -587,7 +536,7 @@ class UI:
         self.interactive_timeout = None
 
     def tick_interactive(self) -> None:
-        """End interactive mode.
+        """Tick timeout of interactive mode.
 
         >>> ui = UI.make()
         >>> ui.tick_interactive()
@@ -646,7 +595,7 @@ class UI:
             return
         forbidden = set(changes) - {
             "duration_mode",
-            "verbose_mode",
+            "query_display_mode",
             "sort_key",
             "query_mode",
             "refresh_time",
@@ -663,8 +612,8 @@ class UI:
         """Return the column matching 'key'.
 
         >>> ui = UI.make()
-        >>> ui.column("cpu")
-        Column(name='CPU%', template_h='%6s ', mandatory=False, sort_key=<SortKey.cpu: 1>)
+        >>> ui.column("cpu")  # doctest: +ELLIPSIS
+        Column(name='CPU%', mandatory=False, sort_key=...)
         >>> ui.column("gloups")
         Traceback (most recent call last):
           ...
@@ -725,7 +674,7 @@ class LoadAverage:
 
 
 @attr.s(auto_attribs=True, frozen=True, slots=True)
-class IOCounter(Deserializable):
+class IOCounter:
     count: int
     bytes: int
     chars: int = 0
@@ -803,7 +752,7 @@ def locktype(value: str) -> LockType:
 class BaseProcess:
     pid: int
     application_name: str
-    database: str
+    database: Optional[str]
     user: str
     client: str
     duration: Optional[float]
@@ -813,16 +762,16 @@ class BaseProcess:
 
 
 @attr.s(auto_attribs=True, frozen=True, slots=True)
-class RunningProcess(BaseProcess, Deserializable):
+class RunningProcess(BaseProcess):
     """Process for a running query."""
 
-    wait: bool
+    wait: Union[bool, None, str]
     is_parallel_worker: bool
 
 
 @attr.s(auto_attribs=True, frozen=True, slots=True)
-class BWProcess(BaseProcess):
-    """Process for a blocking or waiting query."""
+class WaitingProcess(BaseProcess):
+    """Process for a waiting query."""
 
     # Lock information from pg_locks view
     # https://www.postgresql.org/docs/current/view-pg-locks.html
@@ -835,7 +784,22 @@ class BWProcess(BaseProcess):
 
 
 @attr.s(auto_attribs=True, frozen=True, slots=True)
-class SystemProcess(Deserializable):
+class BlockingProcess(BaseProcess):
+    """Process for a blocking query."""
+
+    # Lock information from pg_locks view
+    # https://www.postgresql.org/docs/current/view-pg-locks.html
+    mode: str
+    type: LockType = attr.ib(converter=locktype)
+    relation: str
+    wait: Union[bool, None, str]
+
+    # TODO: update queries to select/compute this column.
+    is_parallel_worker: bool = attr.ib(default=False, init=False)
+
+
+@attr.s(auto_attribs=True, frozen=True, slots=True)
+class SystemProcess:
     meminfo: Tuple[int, ...]
     io_read: IOCounter
     io_write: IOCounter
@@ -880,31 +844,62 @@ class SelectableProcesses:
     >>> w.focused
 
     >>> w.focus_next()
+    True
     >>> w.focused
     456
     >>> w.focus_next()
+    True
+    >>> w.focused
+    123
+    >>> _ = w.focus_next(10)
+    >>> w.focused
+    789
+    >>> _ = w.focus_next()
+    >>> w.focused
+    456
+    >>> _ = w.focus_next()
     >>> w.focused
     123
     >>> w.focus_prev()
+    True
     >>> w.focused
     456
     >>> w.focus_prev()
+    True
     >>> w.focused
     789
     >>> w.focused = 789
-    >>> w.focus_next()
-    >>> w.focused
-    456
-    >>> w.focus_prev()
+    >>> w.focus_next(2)
+    True
     >>> w.focused
     789
+    >>> w.focus_prev()
+    True
+    >>> w.focused
+    123
+    >>> _ = w.focus_prev(3)
+    >>> w.focused
+    456
+    >>> _ = w.focus_last()
+    >>> w.focused
+    789
+    >>> _ = w.focus_first()
+    >>> w.focused
+    456
+    >>> w.focused = 789
+    >>> _ = w.focus_prev(2)
+    >>> w.focused
+    456
+    >>> w.focused = 789
     >>> w.set_items(sorted(w.items))
     >>> w.focused
     789
     >>> w.focus_prev()
+    True
     >>> w.focused
     456
     >>> w.focus_next()
+    True
     >>> w.focused
     789
 
@@ -912,17 +907,26 @@ class SelectableProcesses:
     ([789], 789)
     >>> w.toggle_pin_focused()
     >>> w.focus_next()
+    True
     >>> w.toggle_pin_focused()
     >>> w.selected, w.focused
     ([123, 789], 123)
     >>> w.toggle_pin_focused()
     >>> w.focus_next()
+    True
     >>> w.toggle_pin_focused()
     >>> w.selected, w.focused
     ([456, 789], 456)
+
+    >>> w[1]
+    Proc(pid=456)
+    >>> w[1:3]
+    [Proc(pid=456), Proc(pid=789)]
+
     >>> w.reset()
     >>> w.selected, w.focused
     ([], None)
+
     """
 
     items: List[BaseProcess]
@@ -935,6 +939,19 @@ class SelectableProcesses:
     def __iter__(self) -> Iterator[BaseProcess]:
         return iter(self.items)
 
+    @overload
+    def __getitem__(self, i: int) -> BaseProcess:
+        ...
+
+    @overload
+    def __getitem__(self, s: slice) -> List[BaseProcess]:
+        ...
+
+    def __getitem__(
+        self, val: Union[int, slice]
+    ) -> Union[BaseProcess, List[BaseProcess]]:
+        return self.items[val]
+
     @property
     def selected(self) -> List[int]:
         if self.pinned:
@@ -951,7 +968,7 @@ class SelectableProcesses:
     def set_items(self, new_items: Sequence[BaseProcess]) -> None:
         self.items[:] = list(new_items)
 
-    def _position(self) -> Optional[int]:
+    def position(self) -> Optional[int]:
         if self.focused is None:
             return None
         for idx, proc in enumerate(self.items):
@@ -959,23 +976,40 @@ class SelectableProcesses:
                 return idx
         return None
 
-    def focus_next(self) -> None:
+    def focus_next(self, offset: int = 1) -> bool:
         if not self.items:
-            return
-        idx = self._position()
-        if idx is None:
-            next_idx = 0
-        elif idx == len(self.items) - 1:
+            return False
+        idx = self.position()
+        bottom = len(self.items) - 1
+        if idx is None or (offset == 1 and idx == bottom):
             next_idx = 0
         else:
-            next_idx = idx + 1
+            next_idx = min(bottom, idx + offset)
         self.focused = self.items[next_idx].pid
+        return True
 
-    def focus_prev(self) -> None:
+    def focus_prev(self, offset: int = 1) -> bool:
         if not self.items:
-            return
-        idx = self._position() or 0
-        self.focused = self.items[idx - 1].pid
+            return False
+        idx = self.position() or 0
+        if offset == 1:
+            next_idx = idx - offset
+        else:
+            next_idx = max(idx - offset, 0)
+        self.focused = self.items[next_idx].pid
+        return True
+
+    def focus_first(self) -> bool:
+        if not self.items:
+            return False
+        self.focused = self.items[0].pid
+        return True
+
+    def focus_last(self) -> bool:
+        if not self.items:
+            return False
+        self.focused = self.items[-1].pid
+        return True
 
     def toggle_pin_focused(self) -> None:
         assert self.focused is not None
@@ -986,8 +1020,9 @@ class SelectableProcesses:
 
 
 ActivityStats = Union[
-    Iterable[BWProcess],
+    Iterable[WaitingProcess],
     Iterable[RunningProcess],
-    Tuple[Iterable[BWProcess], SystemInfo],
+    Tuple[Iterable[WaitingProcess], SystemInfo],
+    Tuple[Iterable[BlockingProcess], SystemInfo],
     Tuple[Iterable[LocalRunningProcess], SystemInfo],
 ]
